@@ -22,6 +22,7 @@
 # Version: 27 Jan 2022: Moved mask2runs, runs2mask from math
 # Version: 28 Mar 2022: Added iterate_dtype
 # Version: 01 Jun 2022: Added replace_field
+# Version: 15 Oct 2022: Added masked_index, unmasked_index, find_peaks
 
 
 """
@@ -39,8 +40,8 @@ from .numbers import digit_count
 
 
 __all__ = [
-    'isolate_dtype', 'iterate_dtype', 'map_array', 'mask2runs',
-    'replace_field', 'runs2mask'
+    'find_peaks', 'isolate_dtype', 'iterate_dtype', 'map_array', 'mask2runs',
+    'masked_index', 'replace_field', 'runs2mask', 'unmasked_index'
 ]
 
 
@@ -443,3 +444,246 @@ def runs2mask(runs, n=None):
 
     numpy.cumsum(view, out=view)
     return mask
+
+
+def unmasked_index(index, mask):
+    """
+    Convert `index` in a masked aray into the corresponding index in the
+    original.
+
+    Given an array ``x`` of the same size as `mask`, determine the
+    position of ``x[mask][index]`` in the original array ``x``. Since
+    `mask` may be multi-dimensional but ``x[mask]`` is always raveled,
+    the result may be a scalar or a tuple.
+
+    Parameters
+    ----------
+    index : array-like[int]
+        The index or indices into in the masked (raveled) array. Each
+        element must be in
+        ``[-np.count_nonzero(mask), np.count_nonzero(mask))``.
+    mask : array-like[bool]
+        A boolean mask to determine the indices of the array.
+        Non-boolean mask arrays will be interpreted as booleans, as
+        though with ``mask.astype(bool)``. Scalar arrays will be
+        interpreted as one-element arrays, so scalar indexing 
+        allowed.
+
+    Return
+    ------
+    unmasked_index : int or numpy.ndarray or tuple
+        If `mask` is a scalar or 1D, the result is the same size and
+        shape as `index`. Otherwise, it is a tuple of length
+        ``mask.ndim`` each of whose elements corresponds to a dimension,
+        and it the same size and shape as `index``.
+    """
+    index = numpy.asanyarray(index)
+    mask = numpy.asanyarray(mask)
+
+    if index.size == 0:
+        empty = numpy.array([], dtype=numpy.intp)
+        return (empty,) * mask.ndim if mask.ndim > 1 else empty
+    indices = numpy.argwhere(mask)
+    if mask.ndim > 1:
+        return tuple(indices.T[:, index])
+    indices = indices[index].ravel()
+    return indices if index.ndim else indices.item()
+
+
+def masked_index(index, mask):
+    """
+    Convert N-dimensional `index` into its corresponding location in the
+    masked array, if it is in the masked portion.
+
+    Given an array ``x``, of the same size as `mask`, determine the
+    position of ``x[index]`` in ``x[mask]``. If `index` is not in the
+    masked portion, the corressponding output is set to ``-1``.
+
+    Parameters
+    ----------
+    index : tuple
+        The index or indices into the unmasked multi-dimensional array.
+        A non-tuple (scalar or array) index may be used for 1D masks,
+        will be wrapped in a tuple internally to avoid ambiguity. The
+        number of elements in the tuple must match ``mask.ndim``, and
+        all elements must all broadcast together.
+    mask : array-like[bool]
+        A boolean mask to determine the indices of the array.
+        Non-boolean mask arrays will be interpreted as booleans, as
+        though with ``mask.astype(bool)``. Scalar arrays will be
+        interpreted as one-element arrays, so scalar indexing 
+        allowed.
+
+    Return
+    ------
+    masked_index : int or None
+        If `index` corresponds to a `True` element of `mask`, return its
+        position in the masked result. Otherwise, return `None`.
+    """
+    mask = numpy.asanyarray(mask, dtype=numpy.bool_)
+    if not isinstance(index, tuple):
+        index = (index,)
+    if mask.ndim > 0 and mask.size > 0:
+        if mask.ndim != len(index):
+            raise ValueError('index must have length {}'.format(mask.ndim))
+        # The following two lines are a hack until a new mode is added. See
+        # https://github.com/numpy/numpy/issues/15475#issuecomment-1280056626
+        index = tuple(numpy.asanyarray(i) for i in index)
+        index = tuple(numpy.where(i < 0, i + s, i)
+                            for i, s in zip(index, mask.shape))
+        try:
+            index = numpy.ravel_multi_index(index, mask.shape, mode='raise')
+        except ValueError as e:
+            raise IndexError(str(e))
+    mask = mask.ravel()
+    indices = mask.cumsum(dtype=numpy.intp)
+    indices -= 1
+    return numpy.where(mask[index], indices[index], numpy.intp(-1))
+
+
+def find_peaks(arr, n_peaks=None, mode='value', return_values=False):
+    """
+    Find the locations of the `n_peaks` tallest or leftmost (unfiltered)
+    local maxima of `arr`.
+
+    When ``mode=='value'``, the first peak is located at
+    ``arr.argmax()``. The second peak is the maximum value among the
+    numbers that are not monotonically non-increasing away from the
+    first peak. Successive peaks are returned from remaining portions
+    of the array.
+
+    For a plateau with multiple equal peak elements, only the first is
+    returned. Similarly, if multiple peaks with the same values are
+    found, they will be retreived in order of increasing index.
+
+    Parameters
+    ----------
+    arr : array_like
+        The data to search through. Expected to be 1D. Larger dimensions
+        are raveled to avoid the problem of different numbers of results
+        along a given axis.
+    n_peaks : int or None
+        The number of peaks to search for. None or negative searches for
+        all available peaks. If `n_peaks` local maxima are not found,
+        the result will be shorter than requested.
+    mode : str
+        One of {``'value'``, ``'index'``} (case insensitive):
+
+        - ``'value'``: Return up to the first `n_peaks` tallest peaks,
+          regardless of location.
+        - ``'index'``: Return up to the first `n_peaks` local maxima,
+          traversing the array from left to right, regardless of peak
+          height.
+    return_values : bool
+        Return a second array containing the maximum values at each
+        location. Convenience for ``arr[find_peaks(arr, ...)]``.
+
+    Return
+    ------
+    peak_indices : numpy.ndarray[int]
+        Locations of up to `n_peaks` local maxima. May be shorter than
+        `n_peaks` if there are insufficient local maxima in the data.
+    peak_values : numpy.ndarray
+        Elements of `arr` at `peak_indices`.
+
+    Notes
+    -----
+    To get peaks in reverse order with ``mode='index'``, reverse the
+    input: ``arr[::-1]``.
+
+    To get peaks sorted from left to right with ``mode='value'``, sort
+    the results.
+    """
+    arr = numpy.asanyarray(arr).ravel()
+    n = len(arr)
+    if n_peaks is None:
+        n_peaks = -1
+
+    if not isinstance(mode, str):
+        raise TypeError('mode must be one of the strings {{valid, index}}, '
+                        'not {}'.format(type(mode).__name__))
+    cmode = mode.casefold()
+
+    indices = []
+    if cmode == 'value':
+        mask = numpy.ones(arr.shape, dtype=numpy.bool_)
+        while n and len(indices) != n_peaks:
+            index = unmasked_index(arr[mask].argmax(), mask)
+            indices.append(index)
+            # Find surrounding window
+            start = mask[index::-1].argmin() # First zero before
+            if start != 0:
+                start = index - (start - 1)
+            end = mask[index:].argmin()      # First zero after
+            if end == 0:
+                end = len(arr)
+            else:
+                end += index
+            index -= start
+            diff = numpy.diff(arr[start:end])
+            zero = diff.dtype.type(0)
+            # Find left side start of peak
+            if index > 1:
+                sstart = (diff[index - 1::-1] < zero).argmax()
+                # If index is zero, entire left side is non-increasing
+                if sstart != 0:
+                    # Otherwise, adjust for reversal
+                    sstart = index - sstart
+            else:
+                sstart = 0
+            # Find right side end of peak
+            if index < end - start - 2:
+                send = (diff[index:] > zero).argmax()
+                # If index is zero, entire right side is non-increasing
+                if send != 0:
+                    end = start + index + send + 1
+            start += sstart
+            mask[start:end] = False
+            n -= end - start
+    elif cmode == 'index':
+        subset = arr
+        offset = 0
+        while n and len(indices) != n_peaks:
+            if n == 1:
+                index = 0
+                end = 1
+            else:
+                diff = numpy.diff(subset)
+                zero = diff.dtype.type(0)
+                end = (diff < zero).argmax()
+                if diff[end] >= 0:
+                    # end == 0, since subset is constantly rising, set to last
+                    end = n
+                    index = n - 1
+                else:
+                    # Find beginning of plateau
+                    index = (diff[end::-1] > zero).argmax()
+                    # If index is zero, entire subset is plateau, leave it at 0
+                    if index != 0:
+                        # Otherwise, adjust for reversal
+                        index = end - (index - 1)
+            # Found a peak
+            indices.append(index + offset)
+            # Find end of trough
+            if end < n - 2:
+                end += (diff[end + 1:] > zero).argmax() + 1
+                if diff[end] <= 0:
+                    # Trough never ends
+                    end = n
+                else:
+                    # Select first element of uptrend
+                    end += 1
+            else:
+                end = n
+            # Update subset
+            offset += end
+            n -= end
+            subset = subset[end:]
+    else:
+        raise ValueError('mode must be one of {{valid, index}}, '
+                         'not {}'.format(mode))
+
+    indices = numpy.array(indices, dtype=numpy.intp)
+    if return_values:
+        return indices, arr[indices]
+    return indices
